﻿Imports System.IO
Imports System.Text


Public Class ArchiveUtility
    Public FileOpenDialogFilter As String = "Text file|*.txt;*.csv|*.*|*.*"
    Public FileSaveDialogFilter As String = "Text file|*.txt|csv file|*.csv"


    Public Function SaveStringToFile(ByRef content As String, destFilePath As String,
                                     isShowProgressWindow As Boolean) As String
        Return SaveStringToTextFile(content, destFilePath, isShowProgressWindow)
    End Function


    Private Function SaveStringToTextFile(ByRef content As String, filePath As String,
                                          isShowProgressWindow As Boolean) As String
        Try
            ' Convert the string to a byte array
            Dim contentBytes As Byte() = Encoding.UTF8.GetBytes(content)
            Dim totalBytes As Long = contentBytes.Length
            Dim bufferSize As Integer = 8192 ' 8KB buffer
            Dim totalBytesWritten As Long = 0


            ' Open a file stream for writing
            Using fileStream As New FileStream(filePath, FileMode.Create, FileAccess.Write)
                ' Write the data in chunks

                Dim percentage As Integer
                Dim prevPercentage As Integer = -1


                For offset As Long = 0 To totalBytes - 1 Step bufferSize
                    ' Calculate the size of the current chunk
                    Dim chunkSize As Integer = CInt(Math.Min(bufferSize, totalBytes - offset))
                    Dim buffer(chunkSize - 1) As Byte
                    Array.Copy(contentBytes, offset, buffer, 0, chunkSize)

                    ' Write the chunk to the file
                    fileStream.Write(buffer, 0, chunkSize)
                    totalBytesWritten += chunkSize

                    ' Calculate and report progress
                    If isShowProgressWindow Then
                        percentage = CInt((totalBytesWritten / totalBytes) * 100)
                        If percentage Mod 5 = 0 AndAlso prevPercentage <> percentage Then
                            prevPercentage = percentage
                        Else
                            Application.DoEvents()
                        End If
                    End If

                Next
            End Using


            Return String.Empty

        Catch ex As UnauthorizedAccessException
            Return "No permission to access the file: " & Environment.NewLine & ex.Message

        Catch ex As IOException
            Return "Error reading or writing the file: " & Environment.NewLine & ex.Message

        Catch ex As ArgumentException
            Return "Error: " & Environment.NewLine & ex.Message

        Catch ex As Exception

            Return "An error occurred: " & Environment.NewLine & ex.Message
        End Try
    End Function


    Public Function ReadAllText_From_TextFile(file As String, isShowProgressWindow As Boolean) As String

        Using fs As New FileStream(file, FileMode.Open, FileAccess.Read, FileShare.ReadWrite)

            Using sr As New StreamReader(fs)
                Dim totalBytes As Long = fs.Length
                Dim bytesRead As Long = 0
                Dim buffer As Char() = New Char(1023) {}
                Dim readCount As Integer
                Dim result As New StringBuilder()
                Dim progressPercent As Integer
                Dim prevProgressPercent As Integer = -1
                Using reader As New StreamReader(fs)
                    While reader.Peek() >= 0
                        readCount = reader.Read(buffer, 0, buffer.Length)
                        result.Append(buffer, 0, readCount)
                        bytesRead += readCount

                        If isShowProgressWindow Then
                            progressPercent = CInt((bytesRead / totalBytes) * 100)
                            If prevProgressPercent <> progressPercent AndAlso progressPercent Mod 10 = 0 Then
                                prevProgressPercent = progressPercent
                            End If
                        Else
                            Application.DoEvents()
                        End If
                    End While
                End Using

                Return result.ToString()
            End Using
        End Using
    End Function


    Public Function ReadAllText_From_File(fileName As String, isShowProgressWindow As Boolean) As String
        Return ReadAllText_From_TextFile(fileName, isShowProgressWindow)
    End Function



    Public Function Get_LogFile_String2DAarry_From_TextFile_Parallel(parFilename As String,
                                                                     excludeEntireLineWhenNullCellDataFound As Boolean,
                                                                     isShowProgressWindow As Boolean) As String()()


        Dim bufTable As String()()

        Dim strLines As New List(Of String)()

        Dim totalBytes As Long = New FileInfo(parFilename).Length
        Dim currentBytesRead As Long = 0

        Dim strLine As String
        Dim strLineCount As Integer = 0

        Try
            Dim prevProgressPercent As Integer = -1
            Dim progressPercent As Integer

            Using reader As StreamReader = My.Computer.FileSystem.OpenTextFileReader(parFilename)
                While (True)
                    strLine = reader.ReadLine()
                    If strLine Is Nothing Then Exit While
                    strLines.Add(strLine)
                    strLineCount += 1
                    currentBytesRead += strLine.Length + 2

                    progressPercent = CSng(currentBytesRead) / CSng(totalBytes) * 10

                    If prevProgressPercent <> progressPercent AndAlso progressPercent Mod 2 = 0 Then
                        If isShowProgressWindow Then
                            prevProgressPercent = progressPercent
                        Else
                            Application.DoEvents()
                        End If
                    End If

                End While
            End Using
        Catch ex As Exception
            Return Nothing
        End Try

        Dim delimitStr As String


        If InStr(strLines(0), ",") <> 0 Then
            delimitStr = ","
        Else
            delimitStr = vbTab
        End If

        bufTable = Get_LogFile_String2DAarry_From_ListOfString_Parallel(strLines, delimitStr, isShowProgressWindow)

        Return bufTable
    End Function


    Public Function Get_LogFile_String2DAarry_From_File_Parallel(filename As String,
                                                                 excludeEntireLineWhenNullCellDataFound As Boolean,
                                                                 isShowProgressWindow As Boolean) As String()()

        Return Get_LogFile_String2DAarry_From_TextFile_Parallel(filename,
                                                                excludeEntireLineWhenNullCellDataFound,
                                                                isShowProgressWindow)
    End Function


    'Index begins from 0
    Public Function Get_LogFile_String2DAarry_From_ListOfString_Parallel(
                                                                         rows As List(Of String),
                                                                         delimitChar As String,
                                                                         isShowProgressWindow As Boolean) As String()()

        Const uiUpdateCount As Integer = 5
        Dim matrix As String()() = New String(rows.Count - 1)() {}


        If rows.Count < 5000 Then
            ParallelProgressValue = 0
            Get_LogFile_String2DAarry_From_ListOfString_Parallel_Unit(rows,
                                                                      matrix,
                                                                      delimitChar,
                                                                      0,
                                                                      rows.Count - 1,
                                                                      uiUpdateCount,
                                                                      isShowProgressWindow)
        Else
            ParallelProgressValue = 0

            Parallel.For(1, ParallelCount + 1,
                         New ParallelOptions With {
                            .MaxDegreeOfParallelism = ParallelCount
                            },
                         Sub(ith)
                             Dim y1, y2 As Integer
                             y1 = CInt((matrix.Length / ParallelCount) * (ith - 1)) + 1
                             If ith = 1 Then y1 = 0
                             y2 = Math.Min(CInt((matrix.Length / ParallelCount) * ith), matrix.Length - 1)

                             Get_LogFile_String2DAarry_From_ListOfString_Parallel_Unit(rows,
                                                                                       matrix,
                                                                                       delimitChar,
                                                                                       y1,
                                                                                       y2,
                                                                                       uiUpdateCount,
                                                                                       isShowProgressWindow)
                         End Sub
                         )
        End If


        ReDim Preserve matrix(rows.Count - 2)

        Return matrix
    End Function



    Private Sub Get_LogFile_String2DAarry_From_ListOfString_Parallel_Unit(
                                                                          rows As List(Of String),
                                                                          ByRef matrix As String()(),
                                                                          delimitChar As String,
                                                                          indexBegin As Integer,
                                                                          indexEnd As Integer,
                                                                          uiUpdateCount As Integer,
                                                                          isShowProgressWindow As Boolean)

        Dim curRow As String() = rows(indexBegin).Split(delimitChar)

        Dim indexDate, indexVariable, indexValue As Integer

        'Handling two different data format
        Dim prevDataUpperBound As Integer = -1
        Dim curDataUpperBound As Integer


        Dim uiUpdateInterval As Integer = CInt(CSng(indexEnd - indexBegin) / uiUpdateCount)
        uiUpdateInterval = Math.Max(1, uiUpdateInterval)

        For i As Integer = indexBegin To indexEnd
            curRow = TextLib.SplitAndTrim(rows(i), delimitChar)

            'Handling two different data format
            curDataUpperBound = curRow.GetUpperBound(0)
            If prevDataUpperBound <> curDataUpperBound Then
                If curDataUpperBound = 2 Then
                    indexDate = 0
                    indexVariable = 1
                    indexValue = 2
                ElseIf curDataUpperBound = 4 Then
                    indexDate = 0
                    indexVariable = 3
                    indexValue = 4
                Else
                    matrix(i) = New String(0) {}
                    Continue For
                End If
                prevDataUpperBound = curDataUpperBound
            End If


            'Checking invalid data
            If curRow(indexDate) = "" OrElse curRow(indexVariable) = "" OrElse
               curRow(indexValue) = "" Then

                matrix(i) = New String(0) {}
                Continue For
            End If


            If Date.TryParse(Get_TimeDate(curRow(indexDate)), Nothing) = False Then
                matrix(i) = New String(0) {}
                Continue For
            End If
            If MyText.IsContainSpecialSymbols(curRow(indexVariable),
                                              LoRaPorhibitSpecialSymbols) OrElse
               MyText.IsContainSpecialSymbols(curRow(indexValue),
                                              LoRaPorhibitSpecialSymbols) Then
                matrix(i) = New String(0) {}
                Continue For
            End If


            If i Mod uiUpdateInterval = 0 Then
                If isShowProgressWindow Then
                Else
                    Application.DoEvents()
                End If
            End If


            ParallelProgressValue += 1
            matrix(i) = curRow
        Next
    End Sub

End Class
