﻿Imports System.IO
Imports System.Text

Public Class FileDragDropBox
    Dim WithEvents LocalFile As New FileSystemEngine

    Private ReadOnly lvwColumnSorter As ListViewColumnSorter

    Public Event FileClicked(FileName As String,
                             DirectoryPath As String)

    Public Event FileSelected(FileName As String,
                              DirectoryPath As String)

    Public Event FilesAdded(FileCount As Integer)
    Public Event FileAdding(FileCount As Integer)
    Public Event FileAddingStarted()
    Public Event FileAddingEnded()
    Public Event ProcessingMessage(MsgStr As String)

    Public ValidFileExtension As String = ".csv .txt .zip .7z .tar"
    Public IsByteUnit As Boolean = True
    Public IsStopProcessing As Boolean = False


    Public Function Get_FullPathFileNames() As String()
        Dim FileMaxIndex As Integer = FilesCount - 1
        Dim Filenames(FileMaxIndex) As String

        For q As Integer = 0 To FileMaxIndex
            Filenames(q) = FolderName(q) + Path.DirectorySeparatorChar + FileName(q)
        Next


        Return Filenames
    End Function

    Public ReadOnly Property FilesCount As Integer
        Get
            Return ListView.Items.Count
        End Get
    End Property

    Public ReadOnly Property FileName(FileIndex As Integer) As String
        Get
            With ListView
                If .Items.Count < 1 Then Return ""
                If FileIndex > .Items.Count Then Return ""

                Return .Items(FileIndex).SubItems(0).Text
            End With

            Return ""
        End Get
    End Property


    Public ReadOnly Property FolderName(FileIndex As Integer) As String
        Get
            With ListView
                If .Items.Count < 1 Then Return ""
                If FileIndex > .Items.Count Then Return ""

                Return .Items(FileIndex).SubItems(2).Text

            End With

            Return ""
        End Get
    End Property


    Public Sub Clear()
        ListView.Items.Clear()
    End Sub

    Private Sub FileDragDropBox_Load(sender As Object, e As EventArgs) Handles MyBase.Load
        With ListView
            .Width = Me.Width
            .Height = Me.Height
            .Left = 0
            .Top = 0
            .AllowColumnReorder = False
            .View = View.Details
        End With
    End Sub


    Public Sub StopProcessing()
        IsStopProcessing = True
        LocalFile.StopProcessing()
    End Sub


    Private Sub FileDragDropBox_SizeChanged(sender As Object, e As EventArgs) Handles Me.SizeChanged
        With ListView
            .Width = Me.Width
            .Height = Me.Height
            .Left = 0
            .Top = 0

            ListView.Columns(2).Width = Me.Width -
                                        ListView.Columns(0).Width -
                                        ListView.Columns(1).Width -
                                        25

        End With
    End Sub

    Private Sub ListView_ColumnClick(sender As Object, e As ColumnClickEventArgs) Handles ListView.ColumnClick
        If (e.Column = lvwColumnSorter.SortColumn) Then
            If (lvwColumnSorter.Order = SortOrder.Ascending) Then
                lvwColumnSorter.Order = SortOrder.Descending
            Else
                lvwColumnSorter.Order = SortOrder.Ascending
            End If
        Else
            lvwColumnSorter.SortColumn = e.Column
            lvwColumnSorter.Order = SortOrder.Ascending
        End If
        '
        Me.ListView.Sort()
    End Sub


    Private Sub ListView_DragDrop(sender As Object, e As DragEventArgs) Handles ListView.DragDrop
        If e.Data.GetDataPresent(DataFormats.FileDrop) Then
            Dim MyFiles() As String

            MyFiles = CType(e.Data.GetData(DataFormats.FileDrop), String())

            RaiseEvent FileAddingStarted()

            Call Me.Add(MyFiles)

            RaiseEvent FileAddingEnded()
        End If
    End Sub

    Private Sub ListView_DragEnter(sender As Object, e As DragEventArgs) Handles ListView.DragEnter
        If e.Data.GetDataPresent(DataFormats.FileDrop) Then
            Dim MyFiles() As String

            MyFiles = CType(e.Data.GetData(DataFormats.FileDrop), String())
            If InStr(ValidFileExtension,
                     Path.GetExtension(MyFiles(0).ToLower)) <> 0 Then
                e.Effect = DragDropEffects.Copy
            Else
                e.Effect = DragDropEffects.None
            End If
        Else

            If e.Data.GetDataPresent(DataFormats.Bitmap) Then

                Dim ReturnImage As Image

                ReturnImage = CType(e.Data.GetData(GetType(Bitmap)), Image)

                If ReturnImage.Tag.ToString <> "" Then
                    e.Effect = DragDropEffects.Copy
                End If

            Else
                e.Effect = DragDropEffects.None
            End If
        End If
    End Sub

    Private Sub ListView_KeyUp(sender As Object, e As System.Windows.Forms.KeyEventArgs) Handles ListView.KeyUp
        With ListView
            If .Items.Count < 1 Then Exit Sub

            Try
                If ListView.SelectedIndices(0) > -1 Then
                    RaiseEvent FileClicked(
                        .Items(.SelectedIndices(0)).SubItems(0).Text,
                        .Items(.SelectedIndices(0)).SubItems(2).Text)
                End If
            Catch
            End Try
        End With
    End Sub

    Private Sub ListView_MouseUp(sender As Object, e As MouseEventArgs) Handles ListView.MouseUp
        With ListView
            If .Items.Count < 1 Then Exit Sub

            Try

                If ListView.SelectedIndices(0) > -1 Then
                    RaiseEvent FileClicked(
                        .Items(.SelectedIndices(0)).SubItems(0).Text,
                        .Items(.SelectedIndices(0)).SubItems(2).Text)
                End If
            Catch
            End Try
        End With


        If e.Button = MouseButtons.Right Then
            If ListView.Items.Count = 0 Then Exit Sub
            ContMenu.Show(GetMousePosition)
        End If
    End Sub


    Public Sub Add(FileListArray() As String)
        Dim FileUpBound As Integer
        Dim q, w, t As Integer

        Me.ListView.ListViewItemSorter = Nothing

        IsStopProcessing = False
        ListView.Enabled = False
        ListView.BeginUpdate()


        FileUpBound = FileListArray.GetUpperBound(0)

        For q = 0 To FileUpBound

            If LocalFile.IsFolder(FileListArray(q)) Then
                Continue For
            End If


            If q Mod 100 = 0 Then
                Application.DoEvents()
                If IsStopProcessing Then Exit For
            End If

            AddCurrentFileToDropBox(FileListArray(q))
        Next


        ListView.EndUpdate()
        ListView.Enabled = True
        Me.ListView.ListViewItemSorter = lvwColumnSorter

        RaiseEvent FilesAdded(ListView.Items.Count)
    End Sub


    Sub AddCurrentFileToDropBox(sourceFile As String)
        Dim myItem As ListViewItem
        Dim myItemDetails(2) As String


        If InStr(ValidFileExtension,
                 Path.GetExtension(sourceFile).ToLower) <> 0 Then

            myItemDetails(2) = Path.GetDirectoryName(sourceFile)
            myItemDetails(0) = Path.GetFileName(sourceFile)
            If IsByteUnit Then
                myItemDetails(1) = Format(GetFileSize(sourceFile), "###,##0")
            Else
                myItemDetails(1) = Format(GetFileSize(sourceFile) / 1024, "###,###.0")
            End If


            myItem = New ListViewItem(myItemDetails)
            ListView.Items.Add(myItem)

            Application.DoEvents()
            RaiseEvent FileAdding(ListView.Items.Count)
        End If
    End Sub

    Public Function GetFileSize(MyFilePath As String) As Long
        Dim MyFile As New FileInfo(MyFilePath)
        Dim FileSize As Long = MyFile.Length
        Return FileSize
    End Function


    Public Sub DeleteSelected()
        For x As Integer = ListView.Items.Count - 1 To 0 Step -1
            If ListView.Items(x).Selected = True Then
                ListView.Items(x).Remove()
            End If
        Next
    End Sub

    Private Sub ListView_SelectedIndexChanged(sender As Object, e As EventArgs) Handles ListView.SelectedIndexChanged
        With ListView
            If .Items.Count < 1 OrElse .SelectedIndices.Count = 0 Then Exit Sub

            Try

                If ListView.SelectedIndices(0) > -1 Then
                    RaiseEvent FileSelected(
                        .Items(.SelectedIndices(0)).SubItems(0).Text,
                        .Items(.SelectedIndices(0)).SubItems(2).Text)
                End If
            Catch
            End Try
        End With
    End Sub

    Public Sub New()

        ' This call is required by the designer.
        InitializeComponent()

        lvwColumnSorter = New ListViewColumnSorter()
        Me.ListView.ListViewItemSorter = lvwColumnSorter
    End Sub

    Private Sub LocalFile_ProcessingMessage(MsgStr As String) Handles LocalFile.ProcessingMessage
        RaiseEvent ProcessingMessage(MsgStr)
    End Sub

    Private Sub ContMenu_CopySelectedItems_Click(sender As Object, e As EventArgs) _
        Handles ContMenu_CopySelectedItems.Click
        Dim BufStr As New StringBuilder
        Dim UpperIndex As Integer = ListView.SelectedItems.Count - 1


        For q As Integer = 0 To UpperIndex
            For w As Integer = 0 To 2
                BufStr.Append(ListView.SelectedItems(q).SubItems(w).Text)
                If w = 2 Then
                    BufStr.Append(vbCrLf)
                Else
                    BufStr.Append(vbTab)
                End If
            Next
        Next

        If BufStr.Length = 0 Then
            MsgBox("No content to be copied!", MsgBoxStyle.OkOnly, "Copy")
        Else
            Clipboard.SetText(BufStr.ToString)
        End If
    End Sub

    Private Sub ContMenu_CopyAll_Click(sender As Object, e As EventArgs) Handles ContMenu_CopyAll.Click
        Dim BufStr As New StringBuilder
        Dim UpperIndex As Integer = ListView.Items.Count - 1


        For q As Integer = 0 To UpperIndex
            For w As Integer = 0 To 2
                BufStr.Append(ListView.Items(q).SubItems(w).Text)
                If w = 2 Then
                    BufStr.Append(vbCrLf)
                Else
                    BufStr.Append(vbTab)
                End If
            Next
        Next

        If BufStr.Length = 0 Then
            MsgBox("No content to be copied!", MsgBoxStyle.OkOnly, "Copy")
        Else
            Clipboard.SetText(BufStr.ToString)
        End If
    End Sub
End Class
