﻿Imports System.ComponentModel
Imports System.IO
Imports System.Text
Imports ScottPlot
Imports ScottPlot.Control
Imports ScottPlot.Plottable

Public Class Frm_Chart
    Public DoNotUpdate As Boolean = False
    Public IsBatchProcessing As Boolean = False
    Public LastVariable As String

    Dim MyLastPlot1 As ScatterPlot
    Dim MyLastPlot2 As ScatterPlot
    Dim MyLastPlotMarker As ScatterPlot
    Dim CrossHairMark As Crosshair
    Dim TextOnChart As Text
    Dim HighLightPoint As MarkerPlot

    Dim WindowTitle As String
    Dim ChartTitle As String

    ReadOnly ScatterDotSize As Integer = 5
    ReadOnly ScatterLineWidth As Integer = 1


    Public Structure ChartData
        Dim DateString As String
        Dim DatePoint As Date
        Dim DataString As String
        Dim DataValue As Single
    End Structure


    Public Chart_RawData As New List(Of ChartData)
    Public Chart_Raw_Filtered As New List(Of ChartData)
    Public Chart_CenterCurve As New List(Of ChartData)
    Public Chart_Deviation As New List(Of ChartData)
    Public Chart_Deviation_Filtered As New List(Of ChartData)
    Public Chart_AbsDeviation As New List(Of ChartData)
    Public Chart_AbsDeviation_Filtered As New List(Of ChartData)
    Public Chart_Deviation_PeakValue As New List(Of ChartData)
    Public Chart_Deviation_PeakTimeHr As New List(Of ChartData)
    Public Chart_Deviation_TroughValue As New List(Of ChartData)
    Public Chart_Deviation_TroughTimeHr As New List(Of ChartData)
    Public Chart_Deviation_PeakToPeakAmplitude As New List(Of ChartData)
    Public Chart_CurrentData1 As New List(Of ChartData)
    Public Chart_CurrentData2 As New List(Of ChartData)
    Public Chart_TroughMarker As New List(Of ChartData)
    Public Chart_PeakMarker As New List(Of ChartData)


    Public Sub New()

        ' 디자이너에서 이 호출이 필요합니다.
        InitializeComponent()
    End Sub


    Private Sub Me_Closing(sender As Object, e As CancelEventArgs) Handles MyBase.Closing
        Application.Exit()
    End Sub


    Public Sub Generate_Combo_Period(Source_Combo_Period As ComboBox)
        With Source_Combo_Period.Items
            .Clear()
            .Add("Custom duration")
            .Add("0000:00:01 (1 min)")
            .Add("0000:00:02 (2 min)")
            .Add("0000:00:03 (3 min)")
            .Add("0000:00:04 (4 min)")
            .Add("0000:00:05 (5 min)")
            .Add("0000:00:10 (10 min)")
            .Add("0000:00:20 (20 min)")
            .Add("0000:00:30 (30 min)")
            .Add("0000:01:00 (1 hours)")
            .Add("0000:02:00 (2 hours)")
            .Add("0000:03:00 (3 hours)")
            .Add("0000:04:00 (4 hours)")
            .Add("0000:05:00 (5 hours)")
            .Add("0000:06:00 (6 hours)")
            .Add("0000:12:00 (12 hours)")
            .Add("0001:00:00 (1 day)")
            .Add("0002:00:00 (2 day)")
            .Add("0003:00:00 (3 day)")
            .Add("0004:00:00 (4 day)")
            .Add("0005:00:00 (5 day)")
            .Add("0007:00:00 (1 week)")
            .Add("0014:00:00 (2 weeks)")
            .Add("0030:00:00 (1 month)")
            .Add("0060:00:00 (2 months)")
            .Add("0090:00:00 (3 months)")
            .Add("0120:00:00 (4 months)")
            .Add("0180:00:00 (6 months)")
            .Add("0365:00:00 (1 year)")
            .Add("0730:00:00 (2 year)")
            .Add("1095:00:00 (3 year)")
            .Add("1460:00:00 (4 year)")
            .Add("1825:00:00 (5 year)")
            .Add("3650:00:00 (10 year)")
        End With

        Source_Combo_Period.SelectedIndex = 31
    End Sub

    Private Sub Frm_Chart_Load(sender As Object, e As EventArgs) Handles MyBase.Load
        ParallelCount = Convert.ToInt32(Math.Ceiling((Environment.ProcessorCount * 0.75) * 2.0))

        DoNotUpdate = True

        Date_PlotDate2.Value = New Date(Now.Year, Now.Month, Now.Day, 23, 59, 59)
        Date_PlotDate1.MaxDate = Date_PlotDate2.Value

        Generate_Combo_Period(Combo_Period)


        Timer_ChartRefresh.Enabled = False
        Timer_ChartRefresh.Interval = 10000

        Combo_Raw.SelectedIndex = 0
        Combo_CenterCurve.SelectedIndex = 0
        Combo_Deviation.SelectedIndex = 1
        Combo_AbsDeviation.SelectedIndex = 1


        Combo_Variables.ForeColor = Color.White
        Combo_Variables.BackColor = Color.Blue

        'Call Combo_UpdateMethod_SelectedIndexChanged(Nothing, Nothing)
        DoNotUpdate = False


        With Plot_Chart.Configuration
            .Quality = QualityMode.Low
            .DoubleClickBenchmark = False
            .Zoom = True
            .ScrollWheelZoom = True
            .MiddleClickDragZoom = True
        End With
    End Sub


    Sub SelectAndDrawChart(Optional IsCollectAllChartData As Boolean = False,
                           Optional IsKeepAxes As Boolean = False)
        Dim Minimum_DataCount_ToDetectPeak As Integer = 14


        Chart_RawData.Clear()
        Chart_Raw_Filtered.Clear()
        Chart_CenterCurve.Clear()
        Chart_Deviation.Clear()
        Chart_Deviation_Filtered.Clear()
        Chart_AbsDeviation.Clear()
        Chart_AbsDeviation_Filtered.Clear()
        Chart_Deviation_PeakValue.Clear()
        Chart_Deviation_TroughValue.Clear()
        Chart_Deviation_PeakToPeakAmplitude.Clear()
        Chart_CurrentData1.Clear()
        Chart_CurrentData2.Clear()
        Chart_TroughMarker.Clear()
        Chart_PeakMarker.Clear()

        If Date_PlotDate1.Visible Then
            Chart_RawData = Get_ListChartData_Raw(Date_PlotDate1.Value,
                                                  Date_PlotDate2.Value)
        Else
            Dim UserPeriod_Day As Integer = Val(Mid(Combo_Period.Text, 1, 4))
            Dim UserPeriod_Hour As Integer = Val(Mid(Combo_Period.Text, 6, 2))
            Dim UserPeriod_Min As Integer = Val(Mid(Combo_Period.Text, 9, 2))

            Chart_RawData = Get_ListChartData_Raw(UserPeriod_Day,
                                                  UserPeriod_Hour,
                                                  UserPeriod_Min)
            If Chart_RawData Is Nothing Then
                Exit Sub
            End If
        End If


        If Val(Text_Raw_FilterSize.Text) >= 3 Then
            'Chart_Raw_Filtered = Apply_MedianFilter(Chart_RawData,
            '                                 Val(Text_Raw_FilterSize.Text))
            Chart_Raw_Filtered = Apply_MedianFilter(Chart_RawData,
                                                                  Val(Text_Raw_FilterSize.Text))
        Else
            Chart_Raw_Filtered = Copy_ListChartData(Chart_RawData)
        End If

        If IsCollectAllChartData = False Then
            If Radio_Raw.Checked Then
                Select Case Combo_Raw.SelectedIndex
                    Case 0
                        Chart_CurrentData1 = Copy_ListChartData(Chart_RawData)
                        DrawChart(Plot_Chart, MyLastPlot1, MyLastPlot2, MyLastPlotMarker, Nothing, Nothing,
                                  Nothing, Nothing, Nothing, Nothing, Nothing, CrossHairMark,
                                  HighLightPoint, TextOnChart,
                                  Chart_RawData,
                                  Nothing,
                                  Nothing,
                                  Nothing,
                                  IsKeepAxes)
                    Case 1
                        Chart_CurrentData1 = Copy_ListChartData(Chart_RawData)
                        Chart_CurrentData2 = Copy_ListChartData(Chart_Raw_Filtered)
                        DrawChart(Plot_Chart, MyLastPlot1, MyLastPlot2, MyLastPlotMarker, Nothing, Nothing,
                                  Nothing, Nothing, Nothing, Nothing, Nothing, CrossHairMark,
                                  HighLightPoint, TextOnChart,
                                  Chart_RawData,
                                  Chart_Raw_Filtered,
                                  Nothing,
                                  Nothing,
                                  IsKeepAxes)
                    Case Else
                End Select

                Exit Sub
            End If
        End If


        If CInt(Text_CenterCurve_FilterSize.Text) >= 3 Then
            If Combo_CenterCurve.SelectedIndex = 0 Then
                Chart_CenterCurve = Apply_MedianFilter(Chart_Raw_Filtered,
                                                                     CInt(Text_CenterCurve_FilterSize.Text))
            Else
                Chart_CenterCurve = Apply_MedianFilter(Chart_Raw_Filtered,
                                                                     CInt(Text_CenterCurve_FilterSize.Text))
            End If
        Else
            Chart_CenterCurve = Copy_ListChartData(Chart_Raw_Filtered)
        End If


        If IsCollectAllChartData = False Then
            If Radio_CenterCurve.Checked Then
                Chart_CurrentData1 = Copy_ListChartData(Chart_Raw_Filtered)
                Chart_CurrentData2 = Copy_ListChartData(Chart_CenterCurve)
                DrawChart(Plot_Chart, MyLastPlot1, MyLastPlot2, MyLastPlotMarker, Nothing, Nothing,
                          Nothing, Nothing, Nothing, Nothing, Nothing, CrossHairMark,
                          HighLightPoint, TextOnChart,
                          Chart_Raw_Filtered,
                          Chart_CenterCurve,
                          Nothing,
                          Nothing,
                          IsKeepAxes)
                Exit Sub
            End If
        End If


        Chart_Deviation = Get_ListChartData_Deviation(Chart_Raw_Filtered,
                                                      Chart_CenterCurve,
                                                      Check_DeviationCutOff.Checked,
                                                      Val(Text_DeviationCutOff.Text))

        If Val(Text_Deviation_FilterSize.Text) >= 3 Then
            Chart_Deviation_Filtered = Apply_MedianFilter(Chart_Deviation,
                                                                        Val(Text_Deviation_FilterSize.Text))
        Else
            Chart_Deviation_Filtered = Copy_ListChartData(Chart_Deviation)
        End If


        Dim Temp_Chart_PeakValue As List(Of ChartData) =
                Find_DailyPeaks_Highest(Chart_Deviation_Filtered,
                                        Minimum_DataCount_ToDetectPeak)
        Dim Temp_Chart_TroughValue As List(Of ChartData) =
                Find_DailyPeaks_Lowest(Chart_Deviation_Filtered,
                                       Minimum_DataCount_ToDetectPeak)
        Remove_DuplicateHighestLowestPeaks(Temp_Chart_PeakValue,
                                           Temp_Chart_TroughValue,
                                           Chart_Deviation_PeakValue,
                                           Chart_Deviation_TroughValue,
                                           10)
        Chart_Deviation_PeakValue = Copy_ListChartData(Temp_Chart_PeakValue)
        Chart_Deviation_TroughValue = Copy_ListChartData(Temp_Chart_TroughValue)

        Chart_Deviation_TroughTimeHr = Get_PeakOrTroughTimeHr(Chart_Deviation_TroughValue)
        Chart_Deviation_PeakTimeHr = Get_PeakOrTroughTimeHr(Chart_Deviation_PeakValue)


        Chart_Deviation_PeakToPeakAmplitude = Find_DailyPeakToPeakAmplitude(Chart_Deviation_TroughValue,
                                                                            Chart_Deviation_PeakValue)


        If IsCollectAllChartData = False Then
            If Radio_Deviation.Checked Then
                Select Case Combo_Deviation.SelectedIndex
                    Case 0
                        Chart_CurrentData1 = Copy_ListChartData(Chart_Deviation)
                        DrawChart(Plot_Chart, MyLastPlot1, MyLastPlot2, MyLastPlotMarker, Nothing, Nothing,
                                  Nothing, Nothing, Nothing, Nothing, Nothing, CrossHairMark,
                                  HighLightPoint, TextOnChart,
                                  Chart_Deviation,
                                  Nothing,
                                  Nothing,
                                  Nothing,
                                  IsKeepAxes)
                    Case 1
                        If Check_DeviationPeaks.Checked Then
                            Chart_CurrentData1 = Copy_ListChartData(Chart_Deviation_Filtered)
                            Chart_TroughMarker = Copy_ListChartData(Chart_Deviation_TroughValue)
                            Chart_PeakMarker = Copy_ListChartData(Chart_Deviation_PeakValue)
                            DrawChart(Plot_Chart, MyLastPlot1, MyLastPlot2, MyLastPlotMarker, Nothing, Nothing,
                                      Nothing, Nothing, Nothing, Nothing, Nothing, CrossHairMark,
                                      HighLightPoint, TextOnChart,
                                      Chart_Deviation_Filtered,
                                      Nothing,
                                      Chart_Deviation_PeakValue,
                                      Chart_Deviation_TroughValue,
                                      IsKeepAxes)
                        Else
                            Chart_CurrentData1 = Copy_ListChartData(Chart_Deviation)
                            Chart_CurrentData2 = Copy_ListChartData(Chart_Deviation_Filtered)
                            DrawChart(Plot_Chart, MyLastPlot1, MyLastPlot2, MyLastPlotMarker, Nothing, Nothing,
                                      Nothing, Nothing, Nothing, Nothing, Nothing, CrossHairMark,
                                      HighLightPoint, TextOnChart,
                                      Chart_Deviation,
                                      Chart_Deviation_Filtered,
                                      Nothing,
                                      Nothing,
                                      IsKeepAxes)
                        End If
                    Case Else
                End Select

                Exit Sub
            End If
        End If


        Chart_AbsDeviation = Get_ListChartData_AbsDeviation(Chart_Deviation_Filtered,
                                                            Check_AbsDeviationCutOff.Checked,
                                                            Val(Text_AbsDeviationCutOff.Text))


        If Val(Text_AbsDeviation_FilterSize.Text) >= 3 Then
            Chart_AbsDeviation_Filtered = Apply_MedianFilter(Chart_AbsDeviation,
                                                                           Val(Text_AbsDeviation_FilterSize.Text))
        Else
            Chart_AbsDeviation_Filtered = Copy_ListChartData(Chart_AbsDeviation)
        End If


        Dim Temp_Chart_AbsDevPeaks_Highest As List(Of ChartData) =
                Find_DailyPeaks_Highest(Chart_AbsDeviation_Filtered,
                                        Minimum_DataCount_ToDetectPeak)


        If IsCollectAllChartData = False Then
            If Radio_AbsDeviation.Checked Then
                Select Case Combo_AbsDeviation.SelectedIndex
                    Case 0
                        Chart_CurrentData1 = Copy_ListChartData(Chart_AbsDeviation)
                        DrawChart(Plot_Chart, MyLastPlot1, MyLastPlot2, MyLastPlotMarker, Nothing, Nothing,
                                  Nothing, Nothing, Nothing, Nothing, Nothing, CrossHairMark,
                                  HighLightPoint, TextOnChart,
                                  Chart_AbsDeviation,
                                  Nothing,
                                  Nothing,
                                  Nothing,
                                  IsKeepAxes)
                    Case 1

                        Chart_CurrentData1 = Copy_ListChartData(Chart_AbsDeviation)
                        Chart_CurrentData2 = Copy_ListChartData(Chart_AbsDeviation_Filtered)
                        DrawChart(Plot_Chart, MyLastPlot1, MyLastPlot2, MyLastPlotMarker, Nothing, Nothing,
                                  Nothing, Nothing, Nothing, Nothing, Nothing, CrossHairMark,
                                  HighLightPoint, TextOnChart,
                                  Chart_AbsDeviation,
                                  Chart_AbsDeviation_Filtered,
                                  Nothing,
                                  Nothing,
                                  IsKeepAxes)
                        Exit Sub
                    Case Else
                End Select

                Exit Sub
            End If
        End If
    End Sub


    Function Copy_ListChartData(ByRef Source_ListChartData As List(Of ChartData)) As List(Of ChartData)
        Dim Temp_ListChartData As New List(Of ChartData)

        Temp_ListChartData.AddRange(Source_ListChartData)

        Return Temp_ListChartData
    End Function

    Function Find_DailyPeaks_Highest(ByRef Source_ListChartData As List(Of ChartData),
                                     Minimum_DataCount As Integer) As List(Of ChartData)


        If Source_ListChartData.Count < Minimum_DataCount Then
            Return New List(Of ChartData)
        End If


        Dim Temp_Peaks As New List(Of ChartData)

        Dim CurDay As Integer
        Dim CurDayBeginIndex As Integer
        Dim CurDayEndIndex As Integer


        CurDayBeginIndex = 0
        Do
            CurDay = Source_ListChartData(CurDayBeginIndex).DatePoint.Day
            'If CurDay = 2 Then Stop

            'First find the last index of current day
            CurDayEndIndex = CurDayBeginIndex
            For CurDayIndex As Integer = CurDayBeginIndex + 1 To Source_ListChartData.Count - 1
                If CurDay <> Source_ListChartData(CurDayIndex).DatePoint.Day Then
                    CurDayEndIndex = CurDayIndex - 1
                    Exit For
                End If
            Next


            If CurDayBeginIndex = CurDayEndIndex Then
                If CurDayBeginIndex = Source_ListChartData.Count - 1 Then
                    'return results because it's the end of chart data
                    Return Temp_Peaks
                Else
                    'Because only one data point is found, this doesn't need to proess and keep next search
                    CurDayBeginIndex += 1
                    Continue Do
                End If
            End If


            If CurDayEndIndex - CurDayBeginIndex + 1 < Minimum_DataCount Then
                'Because of too small data count, this doesn't need to proess and keep next search
                CurDayBeginIndex += 1
                Continue Do
            End If


            'Search the highest points of the current day

            Dim DailyMax As Single = -10000000000
            Dim HighestPeak_BeginIndex As Single
            Dim HighestPeak_EndIndex As Single


            'Find the the beginning and end of the highest peaks
            For CurIndex As Integer = CurDayBeginIndex To CurDayEndIndex
                If DailyMax < Source_ListChartData(CurIndex).DataValue Then
                    DailyMax = Source_ListChartData(CurIndex).DataValue
                    HighestPeak_BeginIndex = CurIndex
                    HighestPeak_EndIndex = CurIndex
                    Continue For
                End If

                If DailyMax = Source_ListChartData(CurIndex).DataValue Then
                    HighestPeak_EndIndex = CurIndex
                End If
            Next


            Dim DailyPeak As New ChartData
            If ((HighestPeak_EndIndex - HighestPeak_BeginIndex + 1) Mod 2 = 0) Then
                Dim HighestPeak_CenterIndex As Integer =
                        CInt(Math.Floor((HighestPeak_BeginIndex + HighestPeak_EndIndex) / 2))

                DailyPeak.DataValue = (Source_ListChartData(HighestPeak_CenterIndex).DataValue +
                                       Source_ListChartData(HighestPeak_CenterIndex + 1).DataValue) / 2
                DailyPeak.DatePoint =
                    Date.FromOADate((Source_ListChartData(HighestPeak_CenterIndex).DatePoint.ToOADate +
                                     Source_ListChartData(HighestPeak_CenterIndex + 1).DatePoint.ToOADate) / 2)
                DailyPeak.DataString = Format(DailyPeak.DataValue, "0.####")
                DailyPeak.DateString = DailyPeak.DatePoint.ToString("yyyy-MM-dd HH:mm:ss")

            Else
                Dim HighestPeak_CenterIndex As Integer =
                        CInt((HighestPeak_BeginIndex + HighestPeak_EndIndex) / 2)
                DailyPeak.DataString = Source_ListChartData(HighestPeak_CenterIndex).DataString
                DailyPeak.DataValue = Source_ListChartData(HighestPeak_CenterIndex).DataValue
                DailyPeak.DatePoint = Source_ListChartData(HighestPeak_CenterIndex).DatePoint
                DailyPeak.DateString = Source_ListChartData(HighestPeak_CenterIndex).DateString
            End If


            If HighestPeak_BeginIndex > 0 AndAlso HighestPeak_EndIndex < Source_ListChartData.Count - 1 Then
                If Source_ListChartData(HighestPeak_BeginIndex - 1).DataValue <
                   Source_ListChartData(HighestPeak_BeginIndex).DataValue AndAlso
                   Source_ListChartData(HighestPeak_EndIndex).DataValue >
                   Source_ListChartData(HighestPeak_EndIndex + 1).DataValue Then

                    Temp_Peaks.Add(DailyPeak)
                End If
            End If


            If CurDayEndIndex = Source_ListChartData.Count - 1 Then
                'Because it's the last day, return the final results
                Return Temp_Peaks
            End If

            CurDayBeginIndex = CurDayEndIndex + 1
        Loop

        Return Temp_Peaks
    End Function


    Function Find_DailyPeaks_Lowest(ByRef Source_ListChartData As List(Of ChartData),
                                    Minimum_DataCount As Integer) As List(Of ChartData)

        If Source_ListChartData.Count < Minimum_DataCount Then
            Return New List(Of ChartData)
        End If


        Dim Temp_Peaks As New List(Of ChartData)

        Dim CurDay As Integer
        Dim CurDayBeginIndex As Integer
        Dim CurDayEndIndex As Integer


        CurDayBeginIndex = 0
        Do
            CurDay = Source_ListChartData(CurDayBeginIndex).DatePoint.Day

            'First find the last index of current day
            CurDayEndIndex = CurDayBeginIndex
            For CurDayIndex As Integer = CurDayBeginIndex + 1 To Source_ListChartData.Count - 1
                If CurDay <> Source_ListChartData(CurDayIndex).DatePoint.Day Then
                    CurDayEndIndex = CurDayIndex - 1
                    Exit For
                End If
            Next


            If CurDayBeginIndex = CurDayEndIndex Then
                If CurDayBeginIndex = Source_ListChartData.Count - 1 Then
                    'return results because it's the end of chart data
                    Return Temp_Peaks
                Else
                    'Because only one data point is found, this doesn't need to proess and keep next search
                    CurDayBeginIndex += 1
                    Continue Do
                End If
            End If


            If CurDayEndIndex - CurDayBeginIndex + 1 < Minimum_DataCount Then
                'Because of too small data count, this doesn't need to proess and keep next search
                CurDayBeginIndex += 1
                Continue Do
            End If


            'Search the highest points of the current day

            Dim DailyMin As Single = 10000000000
            Dim LowestPeak_BeginIndex As Single
            Dim LowestPeak_EndIndex As Single


            'Find the the beginning and end of the highest peaks
            For CurIndex As Integer = CurDayBeginIndex To CurDayEndIndex
                If DailyMin > Source_ListChartData(CurIndex).DataValue Then
                    DailyMin = Source_ListChartData(CurIndex).DataValue
                    LowestPeak_BeginIndex = CurIndex
                    LowestPeak_EndIndex = CurIndex
                    Continue For
                End If

                If DailyMin = Source_ListChartData(CurIndex).DataValue Then
                    LowestPeak_EndIndex = CurIndex
                End If
            Next


            Dim DailyPeak As New ChartData
            If ((LowestPeak_EndIndex - LowestPeak_BeginIndex + 1) Mod 2 = 0) Then
                Dim LowestPeak_CenterIndex As Integer =
                        CInt(Math.Floor((LowestPeak_BeginIndex + LowestPeak_EndIndex) / 2))

                DailyPeak.DataValue = (Source_ListChartData(LowestPeak_CenterIndex).DataValue +
                                       Source_ListChartData(LowestPeak_CenterIndex + 1).DataValue) / 2
                DailyPeak.DatePoint = Date.FromOADate((Source_ListChartData(LowestPeak_CenterIndex).DatePoint.ToOADate +
                                                       Source_ListChartData(LowestPeak_CenterIndex + 1).DatePoint.
                                                           ToOADate) / 2)
                DailyPeak.DataString = Format(DailyPeak.DataValue, "0.####")
                DailyPeak.DateString = DailyPeak.DatePoint.ToString("yyyy-MM-dd HH:mm:ss")

            Else
                Dim LowestPeak_CenterIndex As Integer =
                        CInt((LowestPeak_BeginIndex + LowestPeak_EndIndex) / 2)
                DailyPeak.DataString = Source_ListChartData(LowestPeak_CenterIndex).DataString
                DailyPeak.DataValue = Source_ListChartData(LowestPeak_CenterIndex).DataValue
                DailyPeak.DatePoint = Source_ListChartData(LowestPeak_CenterIndex).DatePoint
                DailyPeak.DateString = Source_ListChartData(LowestPeak_CenterIndex).DateString
            End If

            If LowestPeak_BeginIndex > 0 AndAlso LowestPeak_EndIndex < Source_ListChartData.Count - 1 Then
                If Source_ListChartData(LowestPeak_BeginIndex - 1).DataValue >
                   Source_ListChartData(LowestPeak_BeginIndex).DataValue AndAlso
                   Source_ListChartData(LowestPeak_EndIndex).DataValue <
                   Source_ListChartData(LowestPeak_EndIndex + 1).DataValue Then

                    Temp_Peaks.Add(DailyPeak)
                End If
            End If


            If CurDayEndIndex = Source_ListChartData.Count - 1 Then
                'Because it's the last day, return the final results
                Return Temp_Peaks
            End If

            CurDayBeginIndex = CurDayEndIndex + 1
        Loop

        Return Temp_Peaks
    End Function


    Function Find_DailyPeakToPeakAmplitude(ByRef Source_ListChartData_Lowest As List(Of ChartData),
                                           ByRef Source_ListChartData_Highest As List(Of ChartData))


        If Source_ListChartData_Lowest.Count = 0 OrElse Source_ListChartData_Highest.Count = 0 Then
            Return New List(Of ChartData)
        End If


        Dim Temp_PeakAmplitude As New List(Of ChartData)


        'Generate dictionary of lowest peaks
        Dim Dict_Lowest As New Dictionary(Of String, Single)
        Dim DateKey As String
        For q As Integer = 0 To Source_ListChartData_Lowest.Count - 1
            DateKey = Strings.Left(Source_ListChartData_Lowest(q).DateString, 10)
            If Dict_Lowest.Count = 0 Then
                Dict_Lowest.Add(DateKey, Source_ListChartData_Lowest(q).DataValue)
            Else
                If Dict_Lowest.ContainsKey(DateKey) Then
                    If Dict_Lowest(DateKey) > Source_ListChartData_Lowest(q).DataValue Then
                        Dict_Lowest(DateKey) = Source_ListChartData_Lowest(q).DataValue
                    End If
                Else
                    Dict_Lowest.Add(DateKey, Source_ListChartData_Lowest(q).DataValue)
                End If
            End If
        Next

        'Compare with highest peak. If both lowest and highest are found, compute amplitude
        For q As Integer = 0 To Source_ListChartData_Highest.Count - 1
            Dim DailyPeakAmplitude As ChartData

            DateKey = Strings.Left(Source_ListChartData_Highest(q).DateString, 10)
            If Dict_Lowest.ContainsKey(DateKey) Then
                DailyPeakAmplitude.DataValue = Source_ListChartData_Highest(q).DataValue -
                                               Dict_Lowest(DateKey)
                DailyPeakAmplitude.DatePoint = Source_ListChartData_Highest(q).DatePoint
                DailyPeakAmplitude.DataString = Format(DailyPeakAmplitude.DataValue, "0.####")
                DailyPeakAmplitude.DateString = DateKey + " 12:00:00"

                Temp_PeakAmplitude.Add(DailyPeakAmplitude)
            End If
        Next


        Return Temp_PeakAmplitude
    End Function

    Function Get_PeakOrTroughTimeHr(ByRef Source_ListChartData As List(Of ChartData)) As List(Of ChartData)
        Dim Temp_ListChartData As New List(Of ChartData)

        For q As Integer = 0 To Source_ListChartData.Count - 1

            Dim Temp_ChartData As New ChartData With {
                    .DateString = Source_ListChartData(q).DateString,
                    .DatePoint = Source_ListChartData(q).DatePoint,
                    .DataValue = CSng(Source_ListChartData(q).DatePoint.Hour) +
                                 CSng(Source_ListChartData(q).DatePoint.Minute) / 60,
                    .DataString = Format(.DataValue, "0.####")
                    }

            Temp_ListChartData.Add(Temp_ChartData)
        Next


        Return Temp_ListChartData
    End Function


    Public Sub DrawChart(ByRef Source_Plot_Chart As FormsPlot,
                         ByRef Source_MyLastPlot1 As ScatterPlot,
                         ByRef Source_MyLastPlot2 As ScatterPlot,
                         ByRef Source_MyLastPlotMarker As ScatterPlot,
                         ByRef Source_Check_IsYminZero As CheckBox,
                         ByRef Source_Check_SetXrange As CheckBox,
                         ByRef Source_Check_SetYrange As CheckBox,
                         ByRef Source_Text_Xmin As TextBox,
                         ByRef Source_Text_Xmax As TextBox,
                         ByRef Source_Text_Ymin As TextBox,
                         ByRef Source_Text_Ymax As TextBox,
                         ByRef Source_CrossHairMark As Crosshair,
                         ByRef Source_HighLightPoint As MarkerPlot,
                         ByRef Source_TextOnChart As Text,
                         ByRef Source_ListChartData0 As List(Of ChartData),
                         ByRef Source_ListChartData1 As List(Of ChartData),
                         ByRef Source_ListChartData2 As List(Of ChartData),
                         ByRef Source_ListChartData3 As List(Of ChartData),
                         IsKeepAxes As Boolean)


        Source_Plot_Chart.Plot.Clear()
        Source_MyLastPlot1 = Nothing
        Source_MyLastPlot2 = Nothing
        Source_MyLastPlotMarker = Nothing


        If Source_ListChartData0 IsNot Nothing AndAlso Source_ListChartData0.Count > 0 Then
            AddScatterLines(Source_Plot_Chart, Source_MyLastPlot1, Source_MyLastPlot2,
                            Source_ListChartData0, Color.Blue, 1)
        End If

        If Source_ListChartData1 IsNot Nothing AndAlso Source_ListChartData1.Count > 0 Then
            AddScatterLines(Source_Plot_Chart, Source_MyLastPlot1, Source_MyLastPlot2,
                            Source_ListChartData1, Color.Red, 2)
        End If

        If (Source_ListChartData2 IsNot Nothing AndAlso Source_ListChartData2.Count > 0) OrElse
           (Source_ListChartData3 IsNot Nothing AndAlso Source_ListChartData3.Count > 0) Then
            AddMarkers(Source_Plot_Chart, Source_MyLastPlotMarker,
                       Source_ListChartData2, Source_ListChartData3, Color.Red)
        End If


        If (Source_MyLastPlot1 IsNot Nothing AndAlso Source_MyLastPlot1.PointCount > 0) OrElse
           (Source_MyLastPlot2 IsNot Nothing AndAlso Source_MyLastPlot2.PointCount > 0) Then


            Try
                Source_Plot_Chart.Plot.SetAxisLimits()
            Catch
            End Try


            Try
                Source_Plot_Chart.Plot.SetAxisLimits()
            Catch
            End Try

        Else
            Source_Plot_Chart.Plot.Clear()
        End If


        Source_CrossHairMark = Source_Plot_Chart.Plot.AddCrosshair(0, 0)
        With Source_CrossHairMark
            .Color = Color.Green
        End With

        Source_HighLightPoint = Source_Plot_Chart.Plot.AddPoint(0, 0, Color.Green, 10)

        Source_TextOnChart = Source_Plot_Chart.Plot.AddText("*", 0, 0)
        With Source_TextOnChart
            .FontName = "Microsoft Sans Serif"
            .FontSize = 14
            .FontBold = True
            .BackgroundColor = Color.Green
            .BackgroundFill = True
            .Color = Color.White
        End With
    End Sub


    Public Sub AddScatterLines(ByRef Source_Plot_Chart As FormsPlot,
                               ByRef Source_MyLastPlot1 As ScatterPlot,
                               ByRef Source_MyLastPlot2 As ScatterPlot,
                               ByRef Source_ListChartData As List(Of ChartData),
                               SetColor As Color,
                               PlotNum As Integer)
        If Source_ListChartData Is Nothing Then
            Return
        End If


        If Source_ListChartData IsNot Nothing Then

            Dim RowUpperIndex As Integer = Source_ListChartData.Count - 1

            Dim XDate(RowUpperIndex) As Double
            Dim YData(RowUpperIndex) As Double

            For q As Integer = 0 To RowUpperIndex
                XDate(q) = Source_ListChartData(q).DatePoint.ToOADate
                YData(q) = Source_ListChartData(q).DataValue
            Next

            If PlotNum = 1 Then
                Source_MyLastPlot1 = Source_Plot_Chart.Plot.AddScatter(XDate,
                                                                       YData,
                                                                       SetColor,
                                                                       ScatterLineWidth,
                                                                       ScatterDotSize)
            ElseIf PlotNum = 2 Then
                Source_MyLastPlot2 = Source_Plot_Chart.Plot.AddScatter(XDate,
                                                                       YData,
                                                                       SetColor,
                                                                       ScatterLineWidth,
                                                                       ScatterDotSize)
            Else
            End If
            Source_Plot_Chart.Plot.XAxis.DateTimeFormat(True)
        End If
    End Sub


    Public Sub AddMarkers(ByRef Source_Plot_Chart As FormsPlot,
                          ByRef Source_MyLastMarkerPlot As ScatterPlot,
                          ByRef Source_ListChartData1 As List(Of ChartData),
                          ByRef Source_ListChartData2 As List(Of ChartData),
                          SetColor As Color)
        If Source_ListChartData1 Is Nothing AndAlso Source_ListChartData2 Is Nothing Then
            Return
        End If


        Dim DataCount1 As Integer = 0
        Dim DataCount2 As Integer = 0

        If Source_ListChartData1 IsNot Nothing Then
            DataCount1 = Source_ListChartData1.Count
        End If
        If Source_ListChartData2 IsNot Nothing Then
            DataCount2 = Source_ListChartData2.Count
        End If


        Dim RowUpperIndex As Integer = DataCount1 + DataCount2 - 1

        Dim XDate(RowUpperIndex) As Double
        Dim YData(RowUpperIndex) As Double


        For q As Integer = 0 To DataCount1 - 1
            XDate(q) = Source_ListChartData1(q).DatePoint.ToOADate
            YData(q) = Source_ListChartData1(q).DataValue
        Next
        For q As Integer = DataCount1 To RowUpperIndex
            XDate(q) = Source_ListChartData2(q - DataCount1).DatePoint.ToOADate
            YData(q) = Source_ListChartData2(q - DataCount1).DataValue
        Next


        Source_MyLastMarkerPlot =
            Source_Plot_Chart.Plot.AddScatter(XDate, YData, SetColor, 0, 10,
                                              MarkerShape.filledDiamond)

        'Source_Plot_Chart.Plot.XAxis.DateTimeFormat(True)
    End Sub


    Function Get_ListChartData_Raw(UserPeriod_Day As Long,
                                   UserPeriod_Hour As Long,
                                   UserPeriod_Min As Long) As List(Of ChartData)
        Dim Temp_ListChartData As New List(Of ChartData)

        If DataTableList_IsValid = False Then
            Return Nothing
        End If

        Dim Cobmo_Variable As String = Combo_Variables.Text
        DataTableList_Index = VariableListIndexDict(Cobmo_Variable)

        Dim RowUpperIndex As Integer =
                DataTableList(DataTableList_Index).Count - 1


        Dim IsSkipCheckingTime As Boolean = False
        For q As Integer = 0 To RowUpperIndex
            If DataTableList(DataTableList_Index).Item(q)(1).Trim <> "" Then
                Try
                    If IsSkipCheckingTime OrElse
                       IsWithinPeriod(Date.Parse(Get_TimeDate(DataTableList(DataTableList_Index).Item(q)(0))),
                                      Date.Parse(Get_TimeDate(DataTableList(DataTableList_Index).Item(RowUpperIndex)(0))),
                                      UserPeriod_Day,
                                      UserPeriod_Hour,
                                      UserPeriod_Min) Then

                        If IsNumeric(DataTableList(DataTableList_Index).Item(q)(2)) Then
                            Dim Temp_ChartData As New ChartData With {
                                    .DateString = DataTableList(DataTableList_Index).Item(q)(0).Trim,
                                    .DatePoint = Date.Parse(Get_TimeDate(.DateString)),
                                    .DataString = DataTableList(DataTableList_Index).Item(q)(2).Trim,
                                    .DataValue = .DataString
                                    }
                            Temp_ListChartData.Add(Temp_ChartData)

                            IsSkipCheckingTime = True
                        End If
                    End If

                Catch
                End Try
            End If
        Next

        Return Temp_ListChartData
    End Function


    Function Get_ListChartData_Raw(DateFrom As Date, DateEnd As Date) As List(Of ChartData)

        Dim Temp_ListChartData As New List(Of ChartData)


        Dim Cobmo_Variable As String = Combo_Variables.Text
        DataTableList_Index = VariableListIndexDict(Cobmo_Variable)

        Dim RowUpperIndex As Integer =
                DataTableList(DataTableList_Index).Count - 1

        Dim Index_Date, Index_Variable, Index_Value As Integer

        Index_Date = 0
        Index_Variable = 1
        Index_Value = 2

        For q As Integer = 0 To RowUpperIndex
            If DataTableList(DataTableList_Index).Item(q)(Index_Variable).Trim <> "" Then
                Try
                    If MyTime.IsDateBetweenRange(DateFrom, DateEnd,
                                                 Date.Parse(
                                                     Get_TimeDate(DataTableList(DataTableList_Index).Item(q)(Index_Date)))) _
                        Then

                        If IsNumeric(DataTableList(DataTableList_Index).Item(q)(Index_Value)) Then
                            Dim Temp_ChartData As New ChartData With {
                                    .DateString = DataTableList(DataTableList_Index).Item(q)(Index_Date).Trim,
                                    .DatePoint = Date.Parse(Get_TimeDate(.DateString)),
                                    .DataString = DataTableList(DataTableList_Index).Item(q)(Index_Value).Trim,
                                    .DataValue = .DataString
                                    }
                            Temp_ListChartData.Add(Temp_ChartData)
                        End If
                    End If

                Catch
                End Try

            End If

        Next

        Return Temp_ListChartData
    End Function



    Function Apply_MedianFilter(Source_ListChartData As List(Of ChartData),
                                WindowSize As Integer) As List(Of ChartData)

        Dim Median_ListChartData(Source_ListChartData.Count - 1) As ChartData


        Apply_MedianFilter_Unit(Source_ListChartData,
                                                  Median_ListChartData,
                                                  0,
                                                  Source_ListChartData.Count - 1,
                                                  WindowSize)


        Dim Temp_ListChartData As New List(Of ChartData)

        For q As Integer = 0 To Source_ListChartData.Count - 1
            If Median_ListChartData(q).DateString <> "" Then
                Temp_ListChartData.Add(Median_ListChartData(q))
            End If
        Next

        Return Temp_ListChartData
    End Function


    Public Sub Apply_MedianFilter_Unit(Source_ListChartData As List(Of ChartData),
                                        Median_ListChartData() As ChartData,
                                        IndexBegin As Integer,
                                        IndexEnd As Integer,
                                        WindowSize As Integer)



        For q As Integer = IndexBegin To IndexEnd

            Dim numList As New List(Of Double)
            Dim BeginDate As DateTime =
                    Source_ListChartData(q).DatePoint.Subtract(
                        New TimeSpan(CInt((WindowSize - 1) / 2), 0, 0))
            Dim EndDate As DateTime =
                    Source_ListChartData(q).DatePoint.Add(
                        New TimeSpan(CInt((WindowSize - 1) / 2), 0, 0))
            For w As Integer = 0 To Source_ListChartData.Count - 1
                If Source_ListChartData(w).DatePoint >= BeginDate AndAlso
                   Source_ListChartData(w).DatePoint <= EndDate Then
                    numList.Add(Source_ListChartData(w).DataValue)
                End If
                If Source_ListChartData(w).DatePoint > EndDate Then
                    Exit For
                End If
            Next


            If numList.Count >= 3 Then
                Dim MedValue As Single = MyMath.Compute_Median_Of_UnsortedNumbers(numList.ToArray())

                Median_ListChartData(q) = New ChartData With {
                    .DateString = Source_ListChartData(q).DateString,
                    .DatePoint = Source_ListChartData(q).DatePoint,
                    .DataValue = MedValue,
                    .DataString = Format(.DataValue, "0.####")
                    }
            End If
        Next
    End Sub


    Function Get_ListChartData_Deviation(ByRef Source_ListChartRawData As List(Of ChartData),
                                         Source_ListChartMedianFilterData As List(Of ChartData),
                                         IsCutOff As Boolean,
                                         CutOffValue As Integer) As List(Of ChartData)

        Dim Temp_ListChartData As New List(Of ChartData)

        If Source_ListChartRawData.Count = 0 OrElse
           Source_ListChartMedianFilterData.Count = 0 Then Return Temp_ListChartData

        Dim ListChartUpperIndex As Integer =
                Math.Min(Source_ListChartRawData.Count - 1,
                         Source_ListChartMedianFilterData.Count - 1)

        If IsCutOff Then
            For q As Integer = 0 To ListChartUpperIndex
                If Math.Abs(Source_ListChartRawData(q).DataValue -
                            Source_ListChartMedianFilterData(q).DataValue) <= CutOffValue Then
                    Dim Temp_ChartData As New ChartData With {
                            .DateString = Source_ListChartRawData(q).DateString,
                            .DatePoint = Source_ListChartRawData(q).DatePoint,
                            .DataValue = Source_ListChartRawData(q).DataValue -
                                         Source_ListChartMedianFilterData(q).DataValue,
                            .DataString = Format(.DataValue, "0.####")
                            }

                    Temp_ListChartData.Add(Temp_ChartData)
                End If
            Next
        Else
            For q As Integer = 0 To ListChartUpperIndex
                Dim Temp_ChartData As New ChartData With {
                        .DateString = Source_ListChartRawData(q).DateString,
                        .DatePoint = Source_ListChartRawData(q).DatePoint,
                        .DataValue = Source_ListChartRawData(q).DataValue -
                                     Source_ListChartMedianFilterData(q).DataValue,
                        .DataString = Format(.DataValue, "0.####")
                        }
                Temp_ListChartData.Add(Temp_ChartData)
            Next
        End If


        Return Temp_ListChartData
    End Function


    Function Get_ListChartData_AbsDeviation(ByRef Source_ListChartDeviationData As List(Of ChartData),
                                            IsCutOff As Boolean,
                                            CutOffValue As Integer) As List(Of ChartData)


        Dim Temp_ListChartData As New List(Of ChartData)

        If IsCutOff Then
            For q As Integer = 0 To Source_ListChartDeviationData.Count - 1
                If Math.Abs(Source_ListChartDeviationData(q).DataValue) <= CutOffValue Then
                    Dim Temp_ChartData As New ChartData With {
                            .DateString = Source_ListChartDeviationData(q).DateString,
                            .DatePoint = Source_ListChartDeviationData(q).DatePoint,
                            .DataValue = Math.Abs(Source_ListChartDeviationData(q).DataValue),
                            .DataString = Format(.DataValue, "0.####")
                            }
                    Temp_ListChartData.Add(Temp_ChartData)
                End If
            Next

        Else

            For q As Integer = 0 To Source_ListChartDeviationData.Count - 1

                Dim Temp_ChartData As New ChartData With {
                        .DateString = Source_ListChartDeviationData(q).DateString,
                        .DatePoint = Source_ListChartDeviationData(q).DatePoint,
                        .DataValue = Math.Abs(Source_ListChartDeviationData(q).DataValue),
                        .DataString = Format(.DataValue, "0.####")
                        }
                Temp_ListChartData.Add(Temp_ChartData)
            Next
        End If

        Return Temp_ListChartData
    End Function


    Function IsWithinPeriod(BeginDateTime As Date, EndDateTime As Date,
                            UserPeriod_Day As Long,
                            UserPeriod_Hour As Long,
                            UserPeriod_Min As Long) As Boolean


        'Dim PeriodInSec As Long = UserPeriod_Day * 24 * 60 * 60 +
        '                           UserPeriod_Hour * 60 * 60 +
        '                           UserPeriod_Min * 60

        Dim PeriodInSec As Long = UserPeriod_Day * 86400 +
                                  UserPeriod_Hour * 3600 +
                                  UserPeriod_Min * 60


        If DateDiff(DateInterval.Second, BeginDateTime, EndDateTime) <= PeriodInSec Then
            Return True
        Else
            Return False
        End If
    End Function



    Public Function Refresh_Chart(Optional IsCollectAllChartData As Boolean = False,
                                  Optional IsMendatoryRead As Boolean = False,
                                  Optional IsKeepAxes As Boolean = False) As String
        ' IsCollectAllChartData: 숫자 데이터를 추출하기 위해 간혹 사용됨

        If DoNotUpdate Then Return ""

        DoNotUpdate = True


        Dim LogFilename As String = Text_LogFileName.Text
        If LogFilename = "" Then
            DoNotUpdate = False
            Return "The log file name is empty"
        End If


        Dim LastModified As String =
                MyFile.LastModifiedDateStr(LogFilename)
        If LastModified = "" Then
            DataTableList_IsValid = False
            Combo_Variables.Items.Clear()
            Plot_Chart.Plot.Title("")
            DoNotUpdate = False
            Return "The log file not found"
        Else
            DataTableList_IsValid = True
        End If


        If LastLogFileName <> LogFilename OrElse
           LastLogFileDateStr <> LastModified OrElse
           IsMendatoryRead Then


            Label_Msg.Visible = True
            Label_Msg.Refresh()
            If Read_LogFile(LogFilename) = False Then

                Combo_Variables.Items.Clear()
                DoNotUpdate = False
                MsgBox("The log file Is corrupted", MsgBoxStyle.Critical, "Error")
                Label_Msg.Visible = False
                Label_Msg.Refresh()
                Return "The log file Is corrupted"
            End If


            Dim RetMsg As String = Group_DataTable(60, 100, 15)
            If RetMsg <> "" Then
                Combo_Variables.Items.Clear()
                DoNotUpdate = False
                MsgBox(RetMsg, MsgBoxStyle.Critical, "Error")
                Label_Msg.Visible = False
                Label_Msg.Refresh()
                Return RetMsg
            End If


            If Combo_Variables.SelectedIndex = -1 Then
                If Combo_Variables.Items.Count > 0 Then
                    Combo_Variables.SelectedIndex = 0
                End If
            End If

            Label_Msg.Visible = False
            Label_Msg.Refresh()

            LastLogFileName = LogFilename
            LastLogFileDateStr = LastModified
        End If

        If IsBatchProcessing = False Then
            For w As Integer = 0 To Combo_Variables.Items.Count - 1
                If Combo_Variables.Items(w).ToString = LastVariable Then
                    If Combo_Variables.SelectedIndex <> w Then
                        Combo_Variables.SelectedIndex = w
                    End If
                End If
            Next
        End If


        Try
            Dim temp As Integer = CInt(Text_Raw_FilterSize.Text)
        Catch ex As Exception
            Text_Raw_FilterSize.Text = 5
        End Try


        Try
            Dim temp As Integer = CInt(Text_CenterCurve_FilterSize.Text)
        Catch ex As Exception
            Text_CenterCurve_FilterSize.Text = 25
        End Try


        Try
            Dim temp As Integer = CInt(Text_Deviation_FilterSize.Text)
        Catch ex As Exception
            Text_Deviation_FilterSize.Text = 5
        End Try

        Try
            Dim temp As Integer = CInt(Text_AbsDeviation_FilterSize.Text)
        Catch ex As Exception
            Text_AbsDeviation_FilterSize.Text = 5
        End Try

        Try
            Dim temp As Integer = CInt(Text_DeviationCutOff.Text)
        Catch ex As Exception
            Text_DeviationCutOff.Text = 100
        End Try

        Try
            Dim temp As Integer = CInt(Text_AbsDeviationCutOff.Text)
        Catch ex As Exception
            Text_AbsDeviationCutOff.Text = 100
        End Try


        SelectAndDrawChart(IsCollectAllChartData, IsKeepAxes)


        If (MyLastPlot1 IsNot Nothing AndAlso MyLastPlot1.PointCount > 0) OrElse
           (MyLastPlot2 IsNot Nothing AndAlso MyLastPlot2.PointCount > 0) Then
            Plot_Chart.Enabled = True
            Try
                Plot_Chart_AxesChanged(Nothing, Nothing)
                Plot_Chart.Refresh(True)
            Catch
            End Try
        Else
            Try
                Plot_Chart.Plot.Clear()
                Plot_Chart_AxesChanged(Nothing, Nothing)
                Plot_Chart.Refresh(True)
            Catch
            End Try
        End If



        DoNotUpdate = False

        Return ""
    End Function


    'Grouping data for fast charting
    Function Group_DataTable(Progress_BeginPercent As Integer,
                             Progress_EndPercent As Integer,
                             Progress_UpdateCount As Integer) As String

        'Clearing variables
        DataTableList.Clear()
        VariableName.Clear()
        VariableListIndexDict.Clear()


        Dim Index_Date, Index_Variable, Index_Value As Integer
        Dim RowUpperIndex As Integer = DataTable.GetUpperBound(0)
        Dim PrevDataUpperBound As Integer
        Dim CurDataUpperBound As Integer

        For i As Integer = 0 To RowUpperIndex
            If DataTable(i).Length > 1 Then
                CurDataUpperBound = DataTable(i).GetUpperBound(0)
                Exit For
            End If
        Next


        'Handling two different data format
        If CurDataUpperBound = 2 Then
            Index_Date = 0
            Index_Variable = 1
            Index_Value = 2
        ElseIf CurDataUpperBound = 4 Then
            Index_Date = 0
            Index_Variable = 3
            Index_Value = 4
        Else
            Return "Invalid data file"
        End If
        PrevDataUpperBound = CurDataUpperBound

        Dim ProgressBarUpdateInterval As Integer = CSng(RowUpperIndex) / Progress_UpdateCount
        ProgressBarUpdateInterval = Math.Max(1, ProgressBarUpdateInterval)


        For i As Integer = 0 To RowUpperIndex

            CurDataUpperBound = DataTable(i).GetUpperBound(0)
            If PrevDataUpperBound <> CurDataUpperBound Then
                If CurDataUpperBound = 2 Then
                    Index_Date = 0
                    Index_Variable = 1
                    Index_Value = 2
                ElseIf CurDataUpperBound = 4 Then
                    Index_Date = 0
                    Index_Variable = 3
                    Index_Value = 4
                Else
                    Continue For
                End If

                PrevDataUpperBound = CurDataUpperBound
            End If


            If DataTable(i)(Index_Date).Length <> 19 Then
                Continue For
            End If


            If DataTable(i)(Index_Variable).Trim = "" Then
                Continue For
            End If


            'Add new variable

            If VariableName.Contains(DataTable(i)(Index_Variable)) = False Then
                If DataTable(i)(Index_Variable) <> "" Then
                    VariableName.Add(DataTable(i)(Index_Variable))
                    VariableListIndexDict.Add(DataTable(i)(Index_Variable).Trim,
                                              VariableName.Count - 1)
                    DataTableList.Add(New List(Of String()))
                End If
            End If


            'Putting data to DataTableList
            DataTableList_Index = VariableListIndexDict(
                DataTable(i)(Index_Variable))
            DataTableList(DataTableList_Index).Add(New String() {DataTable(i)(Index_Date),
                                                                 DataTable(i)(Index_Variable),
                                                                 DataTable(i)(Index_Value)})


            If i Mod ProgressBarUpdateInterval = 0 Then
                Dim progressPercent As Integer =
                        Progress_BeginPercent +
                        CSng(i) / CSng(RowUpperIndex) * (Progress_EndPercent -
                                                     Progress_BeginPercent)

            End If
        Next

        'Checking Combo_Variables
        Combo_Variables.Sorted = False
        If Combo_Variables.Items.Count >= VariableName.Count Then
            Dim DelCount As Integer = Combo_Variables.Items.Count -
                                      VariableName.Count

            For w As Integer = 1 To DelCount
                Combo_Variables.Items.RemoveAt(VariableName.Count)
            Next
        End If


        'Sorting variables and updating Combo_Variables
        VariableName.Sort(New NaturalComparer)

        Dim KeyStr(VariableName.Count - 1) As String
        VariableName.CopyTo(KeyStr, 0)
        For w As Integer = 0 To VariableName.Count - 1
            If w + 1 > Combo_Variables.Items.Count Then
                Combo_Variables.Items.Add(New ListItem("xxxxx Error xxxxx", Color.Red, Color.White))
            End If
            If Combo_Variables.Items(w).ToString <> KeyStr(w) Then
                Combo_Variables.Items(w) = New ListItem(KeyStr(w),
                                                        Color.Black,
                                                        Color.White)
            End If
        Next

        Return ""
    End Function


    'Return false if file not found
    Function Read_LogFile(FileName As String) As Boolean

        If MyFile.FileExists(FileName) = False Then
            Return False
        End If


        DataTable = MyArchive.Get_LogFile_String2DAarry_From_File_Parallel(FileName, True, True)


        If DataTable Is Nothing Then
            Return False
        Else
            Return True
        End If
    End Function


    Public Sub Combo_Variables_SelectedIndexChanged(sender As Object, e As EventArgs) _
        Handles Combo_Variables.SelectedIndexChanged
        Static previousIndex As Integer = Combo_Variables.SelectedIndex

        If DoNotUpdate AndAlso IsBatchProcessing = False Then Exit Sub


        previousIndex = Combo_Variables.SelectedIndex



        LastVariable = Combo_Variables.Text

        Refresh_Chart()
    End Sub

    Private Sub Cmd_OpenLogFile_Click(sender As Object, e As EventArgs) Handles Cmd_OpenLogFile.Click
        Static lastUsedFolder As String = ""

        Cmd_OpenLogFile.Enabled = False

        With OpenFileDialog
            If lastUsedFolder <> "" Then .InitialDirectory = lastUsedFolder
            .FileName = ""
            .Filter = MyArchive.FileOpenDialogFilter
            .Title = "Open log file"
            .CheckFileExists = True
            .CheckPathExists = True
            .ShowReadOnly = False

            If .ShowDialog = DialogResult.OK Then
                lastUsedFolder = Path.GetDirectoryName(.FileName)
                Text_LogFileName.Text = .FileName
                Refresh_Chart(False, True)
            End If
        End With

        Cmd_OpenLogFile.Enabled = True
    End Sub

    Private Sub Text_LogFileName_TextChanged(sender As Object, e As EventArgs) Handles Text_LogFileName.TextChanged
        MyText.Set_CursorToEndOfTextBox(CType(sender, TextBox))
    End Sub

    Private Sub Cmd_RefreshNow_Click(sender As Object, e As EventArgs) Handles Cmd_RefreshNow.Click
        Cmd_RefreshNow.Enabled = False
        ResetZoom()
        Refresh_Chart(False, False)
        Cmd_RefreshNow.Enabled = True
    End Sub

    Private Sub ResetZoom()
        'AxisX-direction Zoom
        Plot_Chart.Plot.ResetLayout()
    End Sub


    Private Sub Timer_ChartRefresh_Tick(sender As Object, e As EventArgs) Handles Timer_ChartRefresh.Tick
        Refresh_Chart()
    End Sub


    Private Sub Cmd_CopyChart_Click(sender As Object, e As EventArgs) Handles Cmd_CopyChart.Click
        Clipboard.SetImage(Convert_ARGB_To_RGB_Bitmap(Plot_Chart.Plot.GetBitmap(True)))
        MsgBox("Chart copied To clipboard", MsgBoxStyle.Information, "Copy")
    End Sub

    Private Sub Cmd_CopyData_Click(sender As Object, e As EventArgs) Handles Cmd_CopyData.Click
        Dim ChartDataStr As String

        ChartDataStr = Get_ChartData_Double_ForCopy(Chart_CurrentData1,
                                                    Chart_CurrentData2, vbTab, True,
                                                    Date.FromOADate(Plot_Chart.Plot.GetAxisLimits.XMin),
                                                    Date.FromOADate(Plot_Chart.Plot.GetAxisLimits.XMax))


        If ChartDataStr = "" Then
            MsgBox("Raw data Is empty", MsgBoxStyle.Critical, "Copy")
        Else
            Clipboard.SetText(ChartDataStr)
            MsgBox("Raw data copied To clipboard", MsgBoxStyle.Information, "Copy")
        End If
    End Sub


    Function Get_ChartData_ForSave(ByRef Source_ChartData As List(Of ChartData),
                                   DelimitChart As String,
                                   IsConfineDate As Boolean,
                                   Date1 As Date,
                                   Date2 As Date) As String

        Dim builder As New StringBuilder

        For q As Integer = 0 To Source_ChartData.Count - 1
            If IsConfineDate Then
                If MyTime.IsDateBetweenRange(
                    Date1, Date2, Source_ChartData(q).DatePoint) = False Then

                    If Date2 < Source_ChartData(q).DatePoint Then
                        Exit For
                    End If

                    Continue For
                End If
            End If

            builder.Append(Source_ChartData(q).DateString + DelimitChart +
                           Source_ChartData(q).DataString + vbCrLf)
        Next

        Return builder.ToString
    End Function


    Function Get_ChartData_ForCopy(ByRef Source_ChartData As List(Of ChartData),
                                   DelimitStr As String,
                                   IsConfineDate As Boolean,
                                   Date1 As Date,
                                   Date2 As Date) As String

        Dim builder As New StringBuilder


        builder.Append("Date" + DelimitStr + "Time (day)" + DelimitStr + "Value" + vbCrLf)


        For q As Integer = 0 To Source_ChartData.Count - 1
            If IsConfineDate Then
                If MyTime.IsDateBetweenRange(
                    Date1, Date2, Source_ChartData(q).DatePoint) = False Then

                    If Date2 < Source_ChartData(q).DatePoint Then
                        Exit For
                    End If

                    Continue For
                End If
            End If

            builder.Append(Source_ChartData(q).DateString + DelimitStr +
                           Format(Source_ChartData(q).DatePoint.Subtract(Date1).TotalSeconds / 60 / 60 / 24, "0.00000") +
                           DelimitStr +
                           Source_ChartData(q).DataString + vbCrLf)
        Next

        Return builder.ToString
    End Function



    Public Function Get_ChartData_Double_ForCopy(ByRef Source_ChartData1 As List(Of ChartData),
                                                 ByRef Source_ChartData2 As List(Of ChartData),
                                                 DelimitChart As String,
                                                 IsConfineDate As Boolean,
                                                 Date1 As Date,
                                                 Date2 As Date) As String

        If Source_ChartData2.Count = 0 Then
            Return Get_ChartData_ForCopy(Source_ChartData1, DelimitChart,
                                         IsConfineDate, Date1, Date2)
        End If


        Dim MaxIndex As Integer = Math.Max(Source_ChartData1.Count - 1,
                                           Source_ChartData2.Count - 1)
        Dim CurMaxIndex As Integer

        Dim curIndex As Integer
        Dim OutArray(MaxIndex, 5) As String


        curIndex = 0
        For q As Integer = 0 To Source_ChartData1.Count - 1
            If IsConfineDate Then
                If MyTime.IsDateBetweenRange(
                    Date1, Date2, Source_ChartData1(q).DatePoint) = False Then

                    If Date2 < Source_ChartData1(q).DatePoint Then
                        Exit For
                    End If

                    Continue For
                End If
            End If

            OutArray(curIndex, 0) = Source_ChartData1(q).DateString
            OutArray(curIndex, 1) = Format(Source_ChartData1(q).DatePoint.Subtract(Date1).TotalSeconds / 60 / 60 / 24,
                                           "0.00000")
            OutArray(curIndex, 2) = Source_ChartData1(q).DataString
            curIndex += 1
        Next
        CurMaxIndex = curIndex - 1


        curIndex = 0
        For q As Integer = 0 To Source_ChartData2.Count - 1
            If IsConfineDate Then
                If MyTime.IsDateBetweenRange(
                    Date1, Date2, Source_ChartData2(q).DatePoint) = False Then

                    If Date2 < Source_ChartData2(q).DatePoint Then
                        Exit For
                    End If

                    Continue For
                End If
            End If


            OutArray(curIndex, 3) = Source_ChartData2(q).DateString
            OutArray(curIndex, 4) = Format(Source_ChartData2(q).DatePoint.Subtract(Date1).TotalSeconds / 60 / 60 / 24,
                                           "0.00000")
            OutArray(curIndex, 5) = Source_ChartData2(q).DataString
            curIndex += 1
        Next

        CurMaxIndex = Math.Max(CurMaxIndex, curIndex - 1)


        Dim builder As New StringBuilder

        builder.Append("[1] Date" + DelimitChart + "[1] Time (day)" + DelimitChart + "[1] Value" + DelimitChart +
                       "[2] Date" + DelimitChart + "[2] Time (day)" + DelimitChart + "[2] Value" + DelimitChart + vbCrLf)


        For q As Integer = 0 To CurMaxIndex
            builder.Append(OutArray(q, 0) + DelimitChart +
                           OutArray(q, 1) + DelimitChart +
                           OutArray(q, 2) + DelimitChart +
                           OutArray(q, 3) + DelimitChart +
                           OutArray(q, 4) + DelimitChart +
                           OutArray(q, 5) + vbCrLf)
        Next


        Return builder.ToString
    End Function



    Private Sub Combo_Period_SelectedIndexChanged(sender As Object, e As EventArgs) _
        Handles Combo_Period.SelectedIndexChanged
        If Combo_Period.SelectedIndex = 0 Then
            Date_PlotDate1.Visible = True
            Date_PlotDate2.Visible = True
            Label_DatePlotMarkBegin.Visible = True
            Label_DatePlotMarkEnd.Visible = True
        Else
            Date_PlotDate1.Visible = False
            Date_PlotDate2.Visible = False
            Label_DatePlotMarkBegin.Visible = False
            Label_DatePlotMarkEnd.Visible = False
            Refresh_Chart()
        End If
    End Sub


    Private Sub Radio_Raw_CheckedChanged(sender As Object, e As EventArgs) Handles Radio_Raw.CheckedChanged
        If sender.Checked Then
            Refresh_Chart()
        End If
    End Sub

    Private Sub Radio_CenterCurve_CheckedChanged(sender As Object, e As EventArgs) _
        Handles Radio_CenterCurve.CheckedChanged
        If sender.Checked Then
            Refresh_Chart()
        End If
    End Sub


    Private Sub Radio_Deviation_CheckedChanged(sender As Object, e As EventArgs) Handles Radio_Deviation.CheckedChanged
        If sender.Checked Then
            Refresh_Chart()
        End If
    End Sub



    Private Sub Radio_AbsDeviation_CheckedChanged(sender As Object, e As EventArgs) _
        Handles Radio_AbsDeviation.CheckedChanged
        If sender.Checked Then
            Refresh_Chart()
        End If
    End Sub

    Private Sub Combo_Raw_SelectedIndexChanged(sender As Object, e As EventArgs) Handles Combo_Raw.SelectedIndexChanged
        Refresh_Chart()
    End Sub

    Private Sub Combo_CenterCurve_SelectedIndexChanged(sender As Object, e As EventArgs) _
        Handles Combo_CenterCurve.SelectedIndexChanged
        Refresh_Chart()
    End Sub

    Private Sub Combo_Deviation_SelectedIndexChanged(sender As Object, e As EventArgs) _
        Handles Combo_Deviation.SelectedIndexChanged
        If Combo_Deviation.SelectedIndex = 0 Then
            Check_DeviationPeaks.Visible = False
        Else
            Check_DeviationPeaks.Visible = True
        End If

        Refresh_Chart()
    End Sub

    Private Sub Combo_AbsDeviation_SelectedIndexChanged(sender As Object, e As EventArgs) _
        Handles Combo_AbsDeviation.SelectedIndexChanged
        Refresh_Chart()
    End Sub

    Private Sub Check_AbsDeviationCutOff_CheckedChanged(sender As Object, e As EventArgs) _
        Handles Check_AbsDeviationCutOff.CheckedChanged
        Text_AbsDeviationCutOff.Visible = Check_AbsDeviationCutOff.Checked
        Refresh_Chart()
    End Sub



    Private Sub Check_DeviationPeaks_CheckedChanged(sender As Object, e As EventArgs) _
        Handles Check_DeviationPeaks.CheckedChanged
        Refresh_Chart()
    End Sub

    Private Sub Check_DeviationCutOff_CheckedChanged(sender As Object, e As EventArgs) _
        Handles Check_DeviationCutOff.CheckedChanged
        Refresh_Chart()
    End Sub


    Sub Remove_DuplicateHighestLowestPeaks(ByRef Source_HighestPeaks As List(Of ChartData),
                                           ByRef Source_LowestPeaks As List(Of ChartData),
                                           ByRef New_HighestPeaks As List(Of ChartData),
                                           ByRef New_LowestPeaks As List(Of ChartData),
                                           MinimumPeakIntervalHour As Integer)

        For q As Integer = 0 To Source_HighestPeaks.Count - 2
            If IsDuplicatePeaks(Source_HighestPeaks(q).DatePoint,
                                Source_HighestPeaks(q + 1).DatePoint,
                                Source_LowestPeaks,
                                MinimumPeakIntervalHour) Then

                If Source_HighestPeaks(q).DataValue > Source_HighestPeaks(q + 1).DataValue Then
                    New_HighestPeaks.Add(Source_HighestPeaks(q))
                    q += 1
                Else
                    New_HighestPeaks.Add(Source_HighestPeaks(q + 1))
                End If

            Else
                New_HighestPeaks.Add(Source_HighestPeaks(q))
                If q = Source_HighestPeaks.Count - 2 Then
                    New_HighestPeaks.Add(Source_HighestPeaks(q + 1))
                End If
            End If
        Next


        For q As Integer = 0 To Source_LowestPeaks.Count - 2
            If IsDuplicatePeaks(Source_LowestPeaks(q).DatePoint,
                                Source_LowestPeaks(q + 1).DatePoint,
                                Source_HighestPeaks,
                                MinimumPeakIntervalHour) Then

                If Source_LowestPeaks(q).DataValue < Source_LowestPeaks(q + 1).DataValue Then
                    New_LowestPeaks.Add(Source_LowestPeaks(q))
                    q += 1
                Else
                    New_LowestPeaks.Add(Source_LowestPeaks(q + 1))
                End If

            Else
                New_LowestPeaks.Add(Source_LowestPeaks(q))
                If q = Source_LowestPeaks.Count - 2 Then
                    New_LowestPeaks.Add(Source_LowestPeaks(q + 1))
                End If
            End If
        Next
    End Sub

    Function IsDuplicatePeaks(PeakDate1 As Date, PeakDate2 As Date,
                              ByRef OtherPeaks As List(Of ChartData),
                              MinimalPeakIntervalHour As Integer) As Boolean
        For q As Integer = 0 To OtherPeaks.Count - 1
            If (PeakDate1 < OtherPeaks(q).DatePoint AndAlso
                OtherPeaks(q).DatePoint < PeakDate2) OrElse
               Math.Abs(DateDiff(DateInterval.Hour, PeakDate1, PeakDate2)) >= MinimalPeakIntervalHour Then
                Return False
            End If
        Next

        Return True
    End Function


    Function Save_ChartData(ByRef Source_ChartData As List(Of ChartData),
                            DestFilename As String,
                            FilExtension As String,
                            IsConfineDate As Boolean,
                            Optional Date1 As Date = #2022/1/1#,
                            Optional Date2 As Date = #2022/2/1#,
                            Optional isShowProgressWindow As Boolean = True) As String


        Dim ChartDataStr As String

        If Source_ChartData.Count > 0 Then
            ChartDataStr = Get_ChartData_ForSave(Source_ChartData,
                                                 If(FilExtension = "csv", ",", vbTab),
                                                 IsConfineDate, Date1, Date2)
        Else
            ChartDataStr = ""
        End If


        If ChartDataStr = "" Then
            Return "Warning : Chart data is empty"
        Else
            Try

                Dim retStr As String =
                        MyArchive.SaveStringToFile(ChartDataStr, DestFilename, isShowProgressWindow)
                Return retStr

            Catch ee As Exception
                Return ee.Message
            End Try
        End If
    End Function

    Private Sub Cmd_Tools_Click(sender As Object, e As EventArgs) Handles Cmd_Tools.Click
        Frm_Tools.Show()
        Frm_Tools.BringToFront()
    End Sub

    Private Sub Date_PlotDate1_ValueChanged(sender As Object, e As EventArgs) Handles Date_PlotDate1.ValueChanged
        With Date_PlotDate1
            .Value = New Date(.Value.Year, .Value.Month, .Value.Day, .0, 0, 0)
            Date_PlotDate2.MinDate = .Value
        End With
    End Sub

    Private Sub Date_PlotDate2_ValueChanged(sender As Object, e As EventArgs) Handles Date_PlotDate2.ValueChanged
        With Date_PlotDate2
            .Value = New Date(.Value.Year, .Value.Month, .Value.Day, 23, 59, 59)
            Date_PlotDate1.MaxDate = .Value
        End With
    End Sub


    Public Sub Cmd_Reload_Click(sender As Object, e As EventArgs) Handles Cmd_Reload.Click
        Cmd_Reload.Enabled = False
        ResetZoom()
        Refresh_Chart(False, True)
        Cmd_Reload.Enabled = True
    End Sub


    Private Sub Combo_Variables_DrawItem(sender As Object, e As DrawItemEventArgs) Handles Combo_Variables.DrawItem
        DrawItem(sender, e)
    End Sub


    Private Sub Plot_Chart_SizeChanged(sender As Object, e As EventArgs) Handles Plot_Chart.SizeChanged
        ChartTitle = Update_ChartTitle(Plot_Chart, Combo_Variables.Text.Trim)
    End Sub



    Private Sub Cmd_VariableDown_Click(sender As Object, e As EventArgs) Handles Cmd_VariableDown.Click
        If Combo_Variables.SelectedIndex <= Combo_Variables.Items.Count - 2 Then
            Combo_Variables.SelectedIndex += 1
        End If
    End Sub

    Private Sub Cmd_VariableUp_Click(sender As Object, e As EventArgs) Handles Cmd_VariableUp.Click
        If Combo_Variables.SelectedIndex >= 1 Then
            Combo_Variables.SelectedIndex -= 1
        End If
    End Sub


    Public Function Update_ChartTitle(Source_Plot_Chart As FormsPlot, VariableName As String) As String
        Dim ChartTitleStr As String

        If Combo_Variables.Items.Count > 0 Then
            If Chart_CurrentData1.Count > 0 Then

                Dim BeginDate As Date = Date.FromOADate(Source_Plot_Chart.Plot.GetAxisLimits.XMin)
                Dim EndDate As Date = Date.FromOADate(Source_Plot_Chart.Plot.GetAxisLimits.XMax)
                Dim DateDiffInDay As Integer = DateDiff(DateInterval.Day, BeginDate, EndDate)

                ChartTitleStr = "[" + VariableName + "]"
            Else
                ChartTitleStr = "Empty"
            End If
        Else
            ChartTitleStr = "Empty"
        End If

        Source_Plot_Chart.Plot.Title(ChartTitleStr)

        Return ChartTitleStr
    End Function


    Public Sub Plot_Chart_AxesChanged(sender As Object, e As EventArgs) Handles Plot_Chart.AxesChanged

        'If DoNotUpdate = True Then Exit Sub


        Dim BeginDate As Date = Date.FromOADate(Plot_Chart.Plot.GetAxisLimits.XMin)
        Dim EndDate As Date = Date.FromOADate(Plot_Chart.Plot.GetAxisLimits.XMax)
        Dim DateDiffInDay As Integer = DateDiff(DateInterval.Day, BeginDate, EndDate)

        Dim NewMarkerSize As Integer


        If DateDiffInDay <= 2 Then
            NewMarkerSize = 7
        ElseIf DateDiffInDay <= 4 Then
            NewMarkerSize = 6
        ElseIf DateDiffInDay <= 7 Then
            NewMarkerSize = 5
        ElseIf DateDiffInDay <= 14 Then
            NewMarkerSize = 4
        ElseIf DateDiffInDay <= 31 Then
            NewMarkerSize = 3
        ElseIf DateDiffInDay <= 31 * 2 Then
            NewMarkerSize = 2
        Else
            NewMarkerSize = 1
        End If


        If MyLastPlot1 IsNot Nothing Then
            MyLastPlot1.MarkerSize = NewMarkerSize
        End If
        If MyLastPlot2 IsNot Nothing Then
            MyLastPlot2.MarkerSize = NewMarkerSize
        End If


        ChartTitle = Update_ChartTitle(Plot_Chart, Combo_Variables.Text.Trim)
    End Sub

End Class