﻿
Imports System.Text

Public Class TextLib

    Public Shared Function SplitAndTrim(sourceStr As String, delimitStr As String) As String()
        ' Split the source string using the provided delimiter
        Dim splitResult As String() = sourceStr.Split(New String() {delimitStr}, StringSplitOptions.None)

        ' Create a new array with the same length as splitResult to store trimmed strings
        Dim trimmedResult(splitResult.Length - 1) As String

        ' Trim each element and store in the trimmedResult array
        For i As Integer = 0 To splitResult.Length - 1
            trimmedResult(i) = splitResult(i).Trim()
        Next

        Return trimmedResult
    End Function


    'Index begins from 0
    Public Function Get_String2DAarry_From_LongString_Parallel(ByRef SourceStr As String,
                                                               DelimitChar As String,
                                                               IsShowProgressWindow As Boolean) As String()()

        Dim rows As String() = SourceStr.Split(vbLf)
        Dim matrix As String()() = New String(rows.Length - 1)() {}
        Const UI_UpdateCount As Integer = 10
        Dim TotalRecordCount(ParallelCount) As Integer

        If rows.Length < 10000 Then
            ParallelProgressValue = 0
            Get_String2DAarry_From_LongString_Parallel_Unit(TotalRecordCount,
                                                            rows,
                                                            matrix,
                                                            DelimitChar,
                                                            0,
                                                            0,
                                                            rows.Length - 1,
                                                            UI_UpdateCount,
                                                            IsShowProgressWindow)
        Else
            ParallelProgressValue = 0

            Parallel.For(1, ParallelCount + 1,
                         New ParallelOptions With {
                            .MaxDegreeOfParallelism = ParallelCount
                            },
                         Sub(ith)
                             Dim y1, y2 As Integer
                             y1 = CInt((matrix.Length / ParallelCount) * (ith - 1)) + 1
                             If ith = 1 Then y1 = 0
                             y2 = Math.Min(CInt((matrix.Length / ParallelCount) * ith), matrix.Length - 1)

                             Get_String2DAarry_From_LongString_Parallel_Unit(TotalRecordCount,
                                                                             rows,
                                                                             matrix,
                                                                             DelimitChar,
                                                                             ith,
                                                                             y1,
                                                                             y2,
                                                                             UI_UpdateCount,
                                                                             IsShowProgressWindow)
                         End Sub
                         )
        End If


        ReDim Preserve matrix(rows.Length - 2)


        Return matrix
    End Function


    Public Sub Get_String2DAarry_From_LongString_Parallel_Unit(
                                                               ByRef TotalRecordCount() As Integer,
                                                               ByRef rows As String(),
                                                               ByRef matrix As String()(),
                                                               DelimitChar As String,
                                                               ParallelUnit As Integer,
                                                               IndexBegin As Integer,
                                                               IndexEnd As Integer,
                                                               UI_UpdateCount As Integer,
                                                               IsShowProgressWindow As Boolean)

        Dim RowUpperIndex As Integer = rows.GetUpperBound(0)

        Dim UIUpdateInterval As Integer = CInt(CSng(IndexEnd - IndexBegin) / UI_UpdateCount)
        UIUpdateInterval = Math.Max(1, UIUpdateInterval)

        TotalRecordCount(ParallelUnit) = 0


        For i As Integer = IndexBegin To IndexEnd
            matrix(i) = SplitAndTrim(rows(i), DelimitChar)

            If IsShowProgressWindow Then
                TotalRecordCount(ParallelUnit) += 1
                If i Mod UIUpdateInterval = 0 Then
                    Dim progressPercent As Integer =
                            Math.Floor(CSng(TotalRecordCount.Sum) / CSng(RowUpperIndex + 1) * 100)

                End If
            End If
        Next
    End Sub


    Public Sub Set_CursorToEndOfTextBox(ByRef SrcTextBox As TextBox)
        If SrcTextBox.TextLength > 0 Then
            SrcTextBox.SelectionStart = SrcTextBox.Text.Length
            SrcTextBox.ScrollToCaret()
        End If
    End Sub


    Public Function Get_String2DAarry_From_File_Parallel(Par_Filename As String,
                                                         IsShowProgressWindow As Boolean) As String()()


        Dim Buf_String As String
        Dim Buf_Table As String()()

        Buf_String = MyArchive.ReadAllText_From_File(Par_Filename, IsShowProgressWindow)

        If Buf_String Is Nothing Then
            Return Nothing
        End If

        Dim fileExt As String = MyFile.Get_FileExtension_From_FullFileName(Par_Filename).ToLower
        Dim DelimitStr As String


        If InStr(Left(Buf_String, 100), ",") <> 0 Then
            DelimitStr = ","
        Else
            DelimitStr = vbTab
        End If


        Buf_Table = Get_String2DAarry_From_LongString_Parallel(Buf_String,
                                                               DelimitStr,
                                                               IsShowProgressWindow)

        Return Buf_Table
    End Function



    Function IsContainSpecialSymbols(ByRef SourceStr As String, ByRef SpecialSymbols() As Char) As Boolean
        Return SourceStr.IndexOfAny(SpecialSymbols) <> -1
    End Function


    'String2DArray: index starts from 1
    Public Function Convert_String2DYXArray_To_LongString(String2DArray(,) As String,
                                                          delimitStr As String,
                                                          ColumnBeginIndex As Integer,
                                                          RowBeingIndex As Integer) As String
        Dim OutStr As New StringBuilder
        Dim RowCountIndex, ColumnCountIndex As Integer
        Dim q, w As Integer

        RowCountIndex = String2DArray.GetUpperBound(0)
        ColumnCountIndex = String2DArray.GetUpperBound(1)

        For q = RowBeingIndex To RowCountIndex
            For w = ColumnBeginIndex To ColumnCountIndex - 1
                OutStr.Append(String2DArray(q, w) + delimitStr)
            Next
            OutStr.Append(String2DArray(q, ColumnCountIndex) + vbCrLf)
        Next

        Return OutStr.ToString
    End Function
End Class
